use Test::More tests => 1;

BEGIN {
use_ok( 'Statistics::MVA::BayesianLinearDiscrimination' );
}

diag( "Testing Statistics::MVA::BayesianLinearDiscrimination $Statistics::MVA::BayesianLinearDiscrimination::VERSION" );
